use gloo::file::callbacks::read_as_text;
use gloo::file::callbacks::FileReader;
use gloo::file::File;

use serde::Deserialize;
use web_sys::Event;
use web_sys::HtmlInputElement;

use yew::html;
use yew::html::TargetCast;
use yew::Component;
use yew::Context;
use yew::Html;

const CHALLENGES: &[&str] = &[
    "Beginner's Gauntlet",
    "Battlebot Battle Royale",
    "Hero Battle Royale",
    "Kalaw's Challenge",
    "Army of One",
    "Ring-Out Challenge",
    "Flip Fire Gauntlet",
    "Vanishing Maze",
    "Mondo Condo",
    "Birds of Prey",
    "Battlebot Revenge",
    "Mach Speed Melee",
    "Galactic Rumble",
    "Stop and Go",
    "Mecha Madness",
    "Rolling Thunder",
    "Blast from the Past",
    "Bubble Battle",
];

#[derive(Deserialize)]
struct FpSave {
    #[serde(rename = "challengeRank")]
    challenge_rank: Vec<i8>,
}

enum Msg {
    Fail(&'static str),
    Show(String),
    Load(File),
}

enum State {
    Failing(&'static str),
    Loading(FileReader),
    Showing(FpSave),
}

struct App(State);

impl Component for App {
    type Message = Msg;
    type Properties = ();

    fn create(_ctx: &Context<Self>) -> Self {
        Self(State::Failing(
            "Select a Freedom Planet 2 save file to begin!",
        ))
    }

    fn update(&mut self, ctx: &Context<Self>, msg: Self::Message) -> bool {
        self.0 = match msg {
            Msg::Fail(msg) => State::Failing(msg),
            Msg::Load(file) => self.load(ctx, file),
            Msg::Show(data) => self.show(data),
        };

        true
    }

    fn view(&self, ctx: &Context<Self>) -> Html {
        let link = ctx.link();

        html! {
            <div id="wrapper">
                <h1>
                    { "Freedom Planet Challenges Rank Viewer" }
                </h1>

                <div class="preview-media">
                    { self.0.view() }
                </div>

                <label for="file-upload">
                    <p>{ ">> Click to select a file <<" }</p>
                </label>

                <p>
                    { "Based on the " }
                    <a href={ "https://github.com/yewstack/yew/tree/master/examples/file_upload" }>
                        { "Yew File Upload" }
                    </a>
                    { " example." }
                </p>

                <input
                    id="file-upload"
                    type="file"
                    accept="application/json"
                    onchange={ link.callback(Self::on_file_changed) }
                />
            </div>
        }
    }
}

impl State {
    fn view(&self) -> Html {
        match self {
            State::Failing(msg) => html! { <p>{ msg }</p> },
            State::Loading(_) => html! { <p>{ "Loading..." }</p> },
            State::Showing(save) => save.view(),
        }
    }
}

impl FpSave {
    fn view(&self) -> Html {
        let ranks = self
            .challenge_rank
            .iter()
            .skip(1)
            .take(CHALLENGES.len())
            .map(|rank| match rank {
                1 => "C",
                2 => "B",
                3 => "A",
                4 => "S",
                _ => "?",
            });

        let rows = CHALLENGES
            .iter()
            .zip(ranks)
            .map(|(challenge, rank)| {
                html! {
                    <tr>
                        <td>{challenge}</td>
                        <td>{rank}</td>
                    </tr>
                }
            })
            .collect::<Html>();

        html! {
            <table>
                <thead>
                    <tr>
                        <th>{ "Challenge" }</th>
                        <th>{ "Rank" }</th>
                    </tr>
                </thead>
                <tbody>
                    { rows }
                </tbody>
            </table>
        }
    }
}

impl App {
    fn on_file_changed(event: Event) -> Msg {
        let elem: HtmlInputElement = event.target_unchecked_into();

        let Some(files) = elem.files() else {
            return Msg::Fail("No save files were selected");
        };

        if files.length() != 1 {
            return Msg::Fail("Expected a single save file");
        }

        Msg::Load(File::from(files.get(0).unwrap()))
    }

    fn load(&mut self, ctx: &Context<App>, file: File) -> State {
        let link = ctx.link().clone();

        let task = read_as_text(&file, move |res| {
            let message = match res {
                Ok(data) => Msg::Show(data),
                Err(_) => Msg::Fail("Could not read save file"),
            };

            link.send_message(message);
        });

        State::Loading(task)
    }

    fn show(&mut self, data: String) -> State {
        match serde_json::from_str(&data) {
            Ok(save) => State::Showing(save),
            _ => State::Failing("Invalid save file"),
        }
    }
}

fn main() {
    yew::Renderer::<App>::new().render();
}
